package main

import (
	"fmt"
	"os"
	"text/template"
)

type (
	TypeInfo struct {
		Name     string
		TypeName string
	}
)

var header string = `// Generated by gen_types. Don't modify manually!

package core
`

var importFmt string = `
import (
	"fmt"
)
`

var assertTemplate string = `
func Assert{{.Name}}(obj Object, msg string) {{.TypeName}} {
	switch c := obj.(type) {
	case {{.TypeName}}:
		return c
	default:
		if msg == "" {
			msg = fmt.Sprintf("Expected %s, got %s", "{{.Name}}", obj.GetType().ToString(false))
		}
		panic(RT.NewError(msg))
	}
}
`

var ensureTemplate string = `
func Ensure{{.Name}}(args []Object, index int) {{.TypeName}} {
	switch c := args[index].(type) {
	case {{.TypeName}}:
		return c
	default:
		panic(RT.NewArgTypeError(index, c, "{{.Name}}"))
	}
}
`

var infoTemplate string = `
func (x {{.TypeName}}) WithInfo(info *ObjectInfo) Object {
	x.info = info
	return x
}
`

func checkError(err error) {
	if err != nil {
		panic(err)
	}
}

func generateAssertions(types []string) {
	filename := "types_assert_gen.go"
	f, err := os.Create(filename)
	checkError(err)
	defer f.Close()

	var assert = template.Must(template.New("assert").Parse(assertTemplate))
	var ensure = template.Must(template.New("ensure").Parse(ensureTemplate))
	f.WriteString(header)
	f.WriteString(importFmt)
	for _, t := range types {
		typeInfo := TypeInfo{
			Name:     t,
			TypeName: t,
		}
		if t[0] == '*' {
			typeInfo.Name = t[1:]
		}
		assert.Execute(f, typeInfo)
		ensure.Execute(f, typeInfo)
	}
}

func generateInfo(types []string) {
	filename := "types_info_gen.go"
	f, err := os.Create(filename)
	checkError(err)
	defer f.Close()

	var info = template.Must(template.New("info").Parse(infoTemplate))

	f.WriteString(header)
	for _, t := range types {
		typeInfo := TypeInfo{
			Name:     t,
			TypeName: t,
		}
		if t[0] == '*' {
			typeInfo.Name = t[1:]
		}
		info.Execute(f, typeInfo)
	}
}

func main() {
	cmd := os.Args[1]
	switch cmd {
	case "assert":
		generateAssertions(os.Args[2:])
	case "info":
		generateInfo(os.Args[2:])
	default:
		fmt.Println("Unknown command: ", cmd)
	}
}
